package com.chelvc.framework.rocketmq.interceptor;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Objects;

import com.chelvc.framework.base.context.ApplicationContextHolder;
import com.chelvc.framework.base.util.ObjectUtils;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.consumer.DefaultMQPushConsumer;
import org.apache.rocketmq.client.consumer.listener.ConsumeConcurrentlyContext;
import org.apache.rocketmq.client.consumer.listener.ConsumeConcurrentlyStatus;
import org.apache.rocketmq.client.consumer.listener.ConsumeOrderlyContext;
import org.apache.rocketmq.client.consumer.listener.ConsumeOrderlyStatus;
import org.apache.rocketmq.client.consumer.listener.MessageListenerConcurrently;
import org.apache.rocketmq.client.consumer.listener.MessageListenerOrderly;
import org.apache.rocketmq.common.message.MessageExt;
import org.apache.rocketmq.spring.annotation.ConsumeMode;
import org.apache.rocketmq.spring.annotation.RocketMQMessageListener;
import org.apache.rocketmq.spring.core.RocketMQListener;
import org.apache.rocketmq.spring.core.RocketMQPushConsumerLifecycleListener;
import org.apache.rocketmq.spring.support.RocketMQMessageConverter;
import org.springframework.aop.framework.AopProxyUtils;
import org.springframework.core.MethodParameter;
import org.springframework.messaging.converter.MessageConverter;
import org.springframework.messaging.converter.SmartMessageConverter;
import org.springframework.messaging.support.MessageBuilder;

/**
 * 原生消息监听器实现
 *
 * @author Woody
 * @date 2023/5/8
 */
@Slf4j
public class NativeRocketMQListener<T> implements RocketMQPushConsumerLifecycleListener {
    private final Class<?> target = AopProxyUtils.ultimateTargetClass(this);
    private Type type;
    private MethodParameter parameter;
    private MessageConverter converter;
    private DefaultMQPushConsumer consumer;

    /**
     * 初始化消息类型
     */
    private void initializeMessageType() {
        Type parameterized = null;
        Class<?> clazz = this.target;
        while (Objects.nonNull(clazz)) {
            for (Type type : clazz.getGenericInterfaces()) {
                if (type instanceof ParameterizedType
                        && Objects.equals(((ParameterizedType) type).getRawType(), RocketMQListener.class)) {
                    parameterized = type;
                    break;
                }
            }
            clazz = clazz.getSuperclass();
        }
        if (Objects.isNull(parameterized)) {
            this.type = Object.class;
        } else {
            Type[] arguments = ((ParameterizedType) parameterized).getActualTypeArguments();
            this.type = Objects.nonNull(arguments) && arguments.length > 0 ? arguments[0] : Object.class;
        }
    }

    /**
     * 初始化消息处理方法参数
     */
    private void initializeMethodParameter() {
        Class<?> clazz;
        Type type = Objects.requireNonNull(this.type, "Message type has not been initialized");
        if (type instanceof ParameterizedType && this.converter instanceof SmartMessageConverter) {
            clazz = (Class<?>) ((ParameterizedType) type).getRawType();
        } else if (type instanceof Class) {
            clazz = (Class<?>) type;
        } else {
            throw new RuntimeException("parameterType:" + type + " of onMessage method is not supported");
        }
        try {
            this.parameter = new MethodParameter(this.target.getMethod("onMessage", clazz), 0);
        } catch (NoSuchMethodException e) {
            throw new RuntimeException("parameterType:" + type + " of onMessage method is not supported");
        }
    }

    /**
     * 获取消息批量消费数量
     *
     * @return 消息数量
     */
    protected int getBatchConsumeSize() {
        return 1;
    }

    /**
     * 消息转换
     *
     * @param message 消息
     * @return 消息
     */
    @SuppressWarnings("unchecked")
    protected T convert(@NonNull MessageExt message) {
        if (Objects.equals(this.type, MessageExt.class)) {
            return (T) message;
        }

        String str = new String(message.getBody(), StandardCharsets.UTF_8);
        if (Objects.equals(this.type, String.class)) {
            return (T) str;
        }

        // If msgType not string, use objectMapper change it.
        try {
            if (this.type instanceof Class) {
                //if the messageType has not Generic Parameter
                return (T) this.converter.fromMessage(MessageBuilder.withPayload(str).build(), (Class<?>) this.type);
            }
            //if the messageType has Generic Parameter, then use SmartMessageConverter#fromMessage with third
            // parameter "conversionHint".
            //we have validate the MessageConverter is SmartMessageConverter in this#getMethodParameter.
            return (T) ((SmartMessageConverter) this.converter).fromMessage(
                    MessageBuilder.withPayload(str).build(),
                    (Class<?>) ((ParameterizedType) this.type).getRawType(),
                    this.parameter
            );
        } catch (Exception e) {
            throw new RuntimeException("cannot convert message to " + this.type, e);
        }
    }

    /**
     * 消费消息
     *
     * @param message 消息对象
     */
    protected void consume(MessageExt message) {
        Method method = Objects.requireNonNull(this.parameter.getMethod());
        try {
            method.invoke(this, this.convert(message));
        } catch (IllegalAccessException | InvocationTargetException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * 顺序消费消息
     *
     * @param messages 消息列表
     * @param context  上下文对象
     * @return 消费状态
     */
    protected ConsumeOrderlyStatus consume(List<MessageExt> messages, ConsumeOrderlyContext context) {
        messages.forEach(this::consume);
        return ConsumeOrderlyStatus.SUCCESS;
    }

    /**
     * 并发消费消息
     *
     * @param messages 消息列表
     * @param context  上下文对象
     * @return 消费状态
     */
    protected ConsumeConcurrentlyStatus consume(List<MessageExt> messages, ConsumeConcurrentlyContext context) {
        messages.forEach(this::consume);
        return ConsumeConcurrentlyStatus.CONSUME_SUCCESS;
    }

    @Override
    public void prepareStart(DefaultMQPushConsumer consumer) {
        // 初始化消息类型及消费方法参数
        this.consumer = consumer;
        this.consumer.setConsumeMessageBatchMaxSize(this.getBatchConsumeSize());
        this.converter = ApplicationContextHolder.getBean(RocketMQMessageConverter.class).getMessageConverter();
        this.initializeMessageType();
        this.initializeMethodParameter();

        // 重置消息监听器
        RocketMQMessageListener annotation = target.getAnnotation(RocketMQMessageListener.class);
        ConsumeMode mode = ObjectUtils.ifNull(annotation, RocketMQMessageListener::consumeMode);
        if (mode == ConsumeMode.ORDERLY) {
            consumer.setMessageListener((MessageListenerOrderly) (messages, context) -> {
                try {
                    return this.consume(messages, context);
                } catch (Exception e) {
                    log.warn("consume message failed", e);
                    return ConsumeOrderlyStatus.SUSPEND_CURRENT_QUEUE_A_MOMENT;
                } finally {
                    this.consumer.setConsumeMessageBatchMaxSize(this.getBatchConsumeSize());
                }
            });
        } else {
            consumer.setMessageListener((MessageListenerConcurrently) (messages, context) -> {
                try {
                    return this.consume(messages, context);
                } catch (Exception e) {
                    log.warn("consume message failed", e);
                    return ConsumeConcurrentlyStatus.RECONSUME_LATER;
                } finally {
                    this.consumer.setConsumeMessageBatchMaxSize(this.getBatchConsumeSize());
                }
            });
        }
    }
}
